Attribute VB_Name = "ModulePushPull"

'   MODULE PUSHPULL: PUSH AND PULL METHOD IMPLEMENTATION

Private Problem As ClassProblem                 ' object holding problem
Private Result As ClassProblemResult            ' object holding results for Simplex
Private Inc As ClassCounters                    ' object holding counters
Private Definition As ClassProblemDefinition    ' object holding problem definition
Private noteDetail As ClassOutputDetailFile     ' object handling output

Private tableau() As Double                     ' tableau to work on
'''Private iterationCount As Long                  ' iteration counter
Private BVS() As Long                           ' basic variable set
Private D As Long                               ' number of slack/surplus variables
Private U As Long                               ' number of artificial variables
Private M As Long                               ' number of constraints
Private S As Long                               ' number of original variables
Private minMax As Integer                       ' objective indicator: MIN= -1 , MAX= 1
Private k As Long
Private r As Long

Private stepName As String

Private sortedCj As New ClassList

Private degeneracyOccured As Boolean
'
'   Push and Pull module entry point
'
Public Sub ExecutePushPull(onProblem As ClassProblem)

On Error GoTo errorExecutePushPull
    
    Set Problem = onProblem
    Set Result = onProblem.Result(PushPull)
    Set Inc = Result.Counters
    Set sortedCj.Inc = Inc
    Set Definition = onProblem.Definition
    
    Result.StartedWhen = Now
    
    Set noteDetail = Result.OutputDetailFile
    noteDetail.OpenOutputFile

    '   (A) Set dimensions and default values for matrix A, BVS, counters, ...
    InitiatePushPull
    
    '   (B) Convert problem into required form
    DoPreliminaries

    '   (1) Convert inequalities into equalities
    DoStep1

    '   (2) Construct initial tableau containing all slack variables in BVS
    DoStep2
    
    noteDetail.ResultDetailHeader
    noteDetail.SingleLine "Initial tableau", "=", 1, , "="
    noteDetail.PrintTableau tableau, S, D, U, M, BVS
    noteDetail.SingleLine "BVS is " & IIf(CompleteBVS, "", "NOT ") & "complete."
    
'---STEP 3: Generate a complete BVS
    Do While Not CompleteBVS
    
        '   rebuild sorted list based on last tableau row (zero values are excluded!)
        sortedCj.RebuildWith tableau
        
        '   look for new r and k within empty BVS-s
        DoStep3
    
        '   something odd happened: notify and exit method
        Inc.Decis
        If Result.Status <> NotCalculated Then GoTo exitDueToStatusChange
        
        Inc.Decis
        If sortedCj.IsEmpty Then
        
            '   k and r were not found, all possibilities were exhausted: notify and exit loop
            noteDetail.SingleLine "All possible incoming Xj have been used."
            Exit Do
            
        End If
        
        noteDetail.SingleLine "Iteration No.: " & Inc.Counter(cntIterations), "=", 1, , "="
        noteDetail.SingleLine "Step 3: completing BVS (k= " & k & ", r= " & r & ")"
        noteDetail.PrintTableau tableau, S, D, U, M, BVS
        Inc.Decis
        If degeneracyOccured Then
            noteDetail.SingleLine "DEGENERACY OCCURED."
            Result.Degeneracy = "YES"
        End If
        noteDetail.SingleLine "BVS is " & IIf(CompleteBVS, "", "NOT ") & "complete."
    
    Loop
        
'---END STEP 3
    
    '   loop to optimal solution
    noteDetail.SingleLine "Continue with Step 4."
    
    stepName = "4-PUSH"
    
    Do While True
        Inc.Loops
        
        Inc.Decis
        Select Case stepName
        
        Case "4-PUSH"
            
'-----------STEP 4: Push to optimal solution
            
            '   use sorted list, based on last row of tableau, to determine the largest element
            sortedCj.RebuildWith tableau
            
            '   do while positive element exists
            Do While sortedCj.Positive
                Inc.Loops
            
                '   calculate next solution
                DoStep4
        
                '   something odd happened: notify and exit method
                Inc.Decis
                If Result.Status <> NotCalculated Then GoTo exitDueToStatusChange
            
                noteDetail.SingleLine "Iteration No.: " & Inc.Counter(cntIterations), "=", 1, , "="
                noteDetail.SingleLine "Step 4: Push to optimal solution (k= " & k & ", r= " & r & ")"
                noteDetail.PrintTableau tableau, S, D, U, M, BVS
                Inc.Decis
                If degeneracyOccured Then
                    noteDetail.SingleLine "DEGENERACY OCCURED."
                    Result.Degeneracy = "YES"
                End If
                
                '   rebuild list and test for positive element
                sortedCj.RebuildWith tableau
                noteDetail.SingleLine "Zj positive does " & _
                                      IIf(sortedCj.Positive, "", "NOT ") & "exist."
            
            Loop
        
            '   continue with Step 5
            noteDetail.SingleLine "Continue with Step 5."
            stepName = "5-TEST"
        
'-----------END STEP 4
        
        Case "5-TEST"

'-----------STEP 5: Test the iteration results
            noteDetail.SingleLine "Step 5: Test the iteration results:", "=", 1
            noteDetail.SingleLine DoStep5(), , , 1
            
'-----------END STEP 5
        
        Case "6-PULL"

'-----------STEP 6:

            DoStep6
    
            '   something odd happened: notify and exit method
            Inc.Decis
            If Result.Status <> NotCalculated Then GoTo exitDueToStatusChange
            
            noteDetail.SingleLine "Iteration No.: " & Inc.Counter(cntIterations), "=", 1, , "="
            noteDetail.SingleLine "Step 6: Pull to feasible solution (k= " & k & ", r= " & r & ")"
            noteDetail.PrintTableau tableau, S, D, U, M, BVS
            Inc.Decis
            If degeneracyOccured Then
                noteDetail.SingleLine "DEGENERACY OCCURED."
                Result.Degeneracy = "YES"
            End If
            noteDetail.SingleLine "Continue with Step 5."
    
            stepName = "5-TEST"

'-----------END STEP 6
        
        Case "OPTIMAL"
        
            Result.Status = Optimal
            
            Exit Do
        
        End Select
        
    Loop
   
'---evaluate results
    ModuleGauss.CalculateOptimalSolution Result, Definition, tableau, BVS
        
    noteDetail.PrintTestResult
    noteDetail.PrintSummary
    
    noteDetail.CloseOutputFile

    Exit Sub

exitDueToStatusChange:
     
    '   something odd happened: notify and exit execution
    
'''    Result.TotalIterations = iterationCount
'''
    noteDetail.SingleLine "Status has changed: " & Result.StatusName
    noteDetail.SingleLine "Execution stopped.", , 1, 1, "="
    
    noteDetail.CloseOutputFile

    Exit Sub

errorExecutePushPull:

    '   error occured during method execution

    noteDetail.CloseOutputFile

    Err.Raise Err.Number, , "Execute Push and Pull method error." & _
                            NewLine(Err.Description)

    Exit Sub

End Sub

'
'   (A) Initiate Push And Pull: Set dimensions and default values
'
Private Sub InitiatePushPull()

    Dim i As Long, j As Long

    '   (a) initiate global method values
    stopExecuting = False
'''    iterationCount = 0
    Inc.Reset
    Result.Degeneracy = "NO"
    
    '   (b) read Objective, s and m
    Inc.Decis
    minMax = IIf(Definition.Objective = "MAX", -1, 1)
    M = Definition.M
    S = Definition.S

    '   (c) calculate D, U
    D = 0
    U = 0
    
    Inc.Loops M
    Inc.Decis M
    For i = 1 To M
        
        D = IIf(Definition.ConstraintType(i) <> "=", D + 1, D)
    Next i

    '   (d) dimension tableau, set initial values to Definition.A
    
    ReDim tableau(1 To M + 1, 1 To S + D + 1)
    
    Inc.Loops M * S
    For i = 1 To M
        
        For j = 1 To S
            
            tableau(i, j) = Definition.A(i, j)
        Next j
    Next i
        
    '(e) dimension BVS, set initial values to zero
    
    ReDim BVS(1 To M)

End Sub

'
'   (B) Preliminaries: Convert problem into required form
'
Private Sub DoPreliminaries()

    Dim i As Long, j As Long

    '   (a) The problem must be a maximization problem; set A(m+1) row
    
    Inc.Loops S
    Inc.MulDiv 2 * S
    For j = 1 To S
        
        tableau(M + 1, j) = -1 * minMax * Definition.Cj(j)
    Next j

    '   (b) RHS must be non-negative; set A(s+D+1) column
    '   ??? "to avoid the occurence of a possible degeneracy, convert all inequality constr. ???
    '   ??? with RHS=0 into 0 by multiplying each one of them by -1." (last article, page 5) ???
    
    For i = 1 To M
        Inc.Loops
        
        tableau(i, S + D + 1) = Abs(Definition.RHS(i))
        
        Inc.Decis
        If Definition.RHS(i) < 0 Then
            
            For j = 1 To S
                Inc.Loops
                
                tableau(i, j) = -1 * tableau(i, j)
                Inc.MulDiv
            Next j
        End If
    Next i

    '   (c) "All variables must be non-negative." (last article, page 5) CONSTRUCT IT LATER !!!
    '   ??? What does this mean?                                                            ???
    '   ??? How do I implement it?                                                          ???

End Sub

'
'   (1) Step 1: Convert inequalities into equalities
'
Private Sub DoStep1()

    Dim i As Long, j As Long

    '   (a) Set tableau(i,j) value 1 or -1, regarding the relation type of the constraint
    
    j = S
    Inc.Loops M
    For i = 1 To M
        
        Inc.Decis
        If Definition.ConstraintType(i) <> "=" Then
            
            '   is slack or surplus
            
            j = j + 1
            Inc.AddSub
            
            '   slack or surplus, set tableau(i,j)
            Inc.Decis
            tableau(i, j) = IIf(Definition.ConstraintType(i) = "<", 1, -1)
            
            '   if RHS negative, multiply with -1
            Inc.Decis
            If (Definition.RHS(i) < 0) Then tableau(i, j) = -1 * tableau(i, j): Inc.MulDiv
            
        End If
    Next i
    
    '   ??? "The coefficient matrix must have a full row rank, since otherwise either no    ???
    '   ??? solution exists or there are redundand equations." (last article, page 7)       ???
    
End Sub

'
'   (2) Step 2: Construct initial tableau containing all slack variables in BVS
'
Private Sub DoStep2()

    Dim i As Long, j As Long

    '   (a) Set tableau(i,j) value 1 or -1, regarding the relation type of the constraint
    
    j = S
    
    Inc.Loops M
    For i = 1 To M
        
        Inc.Decis
        If Definition.ConstraintType(i) <> "=" Then
        
            '   is slack or surplus
            
            j = j + 1: Inc.AddSub
            
            Inc.Decis
            If (Definition.ConstraintType(i) = "<") And _
               (Definition.RHS(i) >= 0) Then
               
                '   is slack and RHS was non-negative
            
                BVS(i) = j
        
            End If
        End If
    Next i
    
    '   ??? "The coefficient matrix must have a full row rank, since otherwise either no    ???
    '   ??? solution exists or there are redundand equations." (last article, page 7)       ???
    '   !!! If so, notify and STOP                                                          !!!
    
    '   execution has just begun => status should be NotCalculated
    
    Result.Status = NotCalculated

End Sub

'
'   (3) Step 3: Generate a complete BVS
'
Private Sub DoStep3()
    
    Dim modTableau()
    Dim i As Long, j As Long
    
    '   loop while k and r are found or sorted list is empty
    
    Do While Not sortedCj.IsEmpty
        Inc.Loops
        '   (1) Incoming variable is the one with the largest Cj
            
        k = sortedCj.Largest.j

        '   (2) Choose the smallest !!positive!! C/R if possible
        '   ??? "If there are alternatives, break the ties arbitrarily." (last article, pg. 7)  ???
        '   ??? What does that mean? How do I implement it? => checking for degeneracy.         ???
            
        '   create modified tableau (clear rows with occupied bvs)
        modTableau = ModuleUtilities.CloneMatrix(tableau)
        For i = 1 To M
            If BVS(i) <> 0 Then
                For j = 1 To S
                    modTableau(i, j) = 0
                Next j
            End If
        Next i
                
        r = ModuleGauss.ColumnRatioTest(Inc, modTableau, k, degeneracyOccured, True)


        Inc.Decis
        If r = 0 Then
            
            '   column ratio test didn't give any results, try with next biggest element in sorted list
            
            sortedCj.RemoveLargest
                
        Else
        
            '   (3a) Calculated r-row in BVS is empty, let's pivot
            Inc.Decis
            If BVS(r) = 0 Then
            
                '   increase iteration count and check for boundaries (to prevent infinite loop)
                Inc.Decis
                If ModuleGauss.IterationCountExceedsMaxAllowed(Inc) Then
    
                    '   too many iterations: change status and leave this loop (to prevent infinite loop)
                    
                    Result.Status = MaxIterationsExceeded
                    Exit Sub
    
                End If
    
                '   success: k and r are found, execute tableau pivoting and leave this loop
                    
                ModuleGauss.GaussPivot Inc, tableau, k, r, BVS
                Exit Do
            
            Else
                    
                '   (3b) BVS(r) is already occupied, repeat (1) with next largest Cj
                
                sortedCj.RemoveLargest
                
            End If
        End If
    Loop
End Sub

'
'   (4) Step 4: Push to optimal solution
'
Private Sub DoStep4()

    '   (1) Incoming variable is the one with the largest Cj
    
    k = sortedCj.Largest.j

    '   (2) Choose the smallest !!positive!! C/R if possible
        
    r = ModuleGauss.ColumnRatioTest(Inc, tableau, k, degeneracyOccured)
    
    Inc.Decis
    If r = 0 Then
    
        '   column ratio test didn't give any result => is it infesible or unbounded?
        
        Result.Status = Unbounded
        Exit Sub
    
    End If

    Inc.Decis
    If ModuleGauss.IterationCountExceedsMaxAllowed(Inc) Then
    
        '   too many iterations: change status and leave this loop (to prevent infinite loop)
                    
        Result.Status = MaxIterationsExceeded
        Exit Sub
    
    End If
            
    '   success: calculate new tableau
    
    ModuleGauss.GaussPivot Inc, tableau, k, r, BVS
            
End Sub

'
'   (5) Step 5: Testing
'
Private Function DoStep5() As String

    Dim decision As String
    
    '   use sorted list to determine existance of positive Cj
    
    sortedCj.RebuildWith tableau
    
    '   decide which step goes next
    
    decision = ""
    decision = decision & IIf(negativeRHS, "F", "T"): Inc.Decis
    decision = decision & IIf(sortedCj.Positive, "F", "T"): Inc.Decis
    
    Inc.Decis
    Select Case decision
    
    Case "TT"
    
        '   All RHS >= 0 and all Cj <= 0 : solution is optimal
        
        stepName = "OPTIMAL"
        DoStep5 = "All RHS >= 0 and all Zj <= 0 : solution is optimal"
    
    Case "TF", "FF"
            
        '   Not all Cj <= 0: continue with PUSH (Step 4)
        
        stepName = "4-PUSH"
        DoStep5 = "Not all Zj <= 0: continue with PUSH (Step 4)"
    
    Case "FT"
        
        '   Not all RHS >= 0: continue with PULL (Step 6)
        
        stepName = "6-PULL"
        DoStep5 = "Not all RHS >= 0: continue with PULL (Step 6)"
    
    End Select

End Function

'
'   (6) Step 6: Pull to feasible solution
'
Private Sub DoStep6()

    '   (a) Use the dual simplex pivoting rules to determine k and r (R/R test)
    
    ModuleGauss.RowRatioTest Inc, tableau, k, r
    
    Inc.Decis
    If r = 0 Then
    
        '   row ratio test didn't give any result => is it infesible or unbounded?
        
        Result.Status = Infeasible
        Exit Sub
    
    End If
    
    Inc.Decis
    If ModuleGauss.IterationCountExceedsMaxAllowed(Inc) Then
    
        '   too many iterations: change status and leave this loop (to prevent infinite loop)
                    
        Result.Status = MaxIterationsExceeded
        Exit Sub
    
    End If
    
    '   (b) generate the next tableau
    
    ModuleGauss.GaussPivot Inc, tableau, k, r, BVS
    
End Sub

'
'   check if BVS is complete
'
Private Function CompleteBVS() As Boolean
    
    Dim i As Long
    
    CompleteBVS = True
    
    For i = 1 To M
        Inc.Loops
    
        Inc.Decis
        If BVS(i) = 0 Then
            
            CompleteBVS = False
            
            Exit Function
        
        End If
    Next i
End Function

'
'   check if RHS negative exists (pull phase)
'
Private Function negativeRHS() As Boolean

    Dim i As Long
    
    negativeRHS = False
    
    For i = 1 To M
        Inc.Loops
        
        Inc.Decis
        If tableau(i, S + D + 1) < 0 Then
            
            negativeRHS = True
            
            Exit Function
        
        End If
     Next i
End Function
