Attribute VB_Name = "ModuleSimplex"

'   MODULE SIMPLEX: STD. SIMPLEX METHOD IMPLEMENTATION

Option Explicit

Private Problem As ClassProblem                 ' object holding problem
Private Result As ClassProblemResult            ' object holding results for Simplex
Private Inc As ClassCounters                    ' object holding counters
Private Definition As ClassProblemDefinition    ' object holding problem definition
Private noteDetail As ClassOutputDetailFile     ' object handling output

Private tableau() As Double                     ' tableau to work on
'''Private iterationCount As Long                  ' iteration counter
Private BVS() As Long                           ' basic variable set
Private D As Long                               ' number of slack/surplus variables
Private U As Long                               ' number of artificial variables
Private M As Long                               ' number of constraints
Private S As Long                               ' number of original variables
Private minMax As Integer                       ' objective indicator: MIN= -1 , MAX= 1
Private bigM As Double                          ' the big M value
Private k As Long
Private r As Long

Private stopExecuting As Boolean                ' execution flag

Private simplexCj() As Double                   ' last row in tableau - why do I need that ??

Private existsZjPositive As Boolean             ' positive Zj - Cj existance flag
Private biggestZjColumnIndex As Long            ' biggest Zj - Cj column index (next possible k)

Private degeneracyOccured As Boolean
'
'   entry point for standardSimplex method execution
'
Public Sub ExecuteSimplex(onProblem As ClassProblem)

On Error GoTo errorExecuteSimplex
    
    Set Problem = onProblem
    Set Result = onProblem.Result(Simplex)
    Set Inc = Result.Counters
    Set Definition = onProblem.Definition
    
    Result.StartedWhen = Now
    
    
    Set noteDetail = Result.OutputDetailFile
    noteDetail.OpenOutputFile
    
    '   (A) set dimensions and default values for matrix A, BVS, counters, ...
    InitiateSimplex

    '   (B) construct initial tableau
    ConstructInitialTableau
    
    noteDetail.ResultDetailHeader
    noteDetail.SingleLine "Initial tableau", "=", 1, , "="
    noteDetail.PrintTableau tableau, S, D, U, M, BVS
    noteDetail.SingleLine "Zj positive does " & IIf(existsZjPositive, "", "NOT ") & "exist."
   
    '   loop to optimal solution
    Do While stopExecuting = False
        Inc.Loops
        '   (1) Calculate new tableau
        CalculateNewTableau
    
        noteDetail.SingleLine "Iteration No.: " & Inc.Counter(cntIterations), "=", 1, , "="
        noteDetail.SingleLine "k= " & k & ", r= " & r
        noteDetail.PrintTableau tableau, S, D, U, M, BVS
        Inc.Decis
        If degeneracyOccured Then
            noteDetail.SingleLine "DEGENERACY OCCURED."
            Result.Degeneracy = "YES"
        End If
        noteDetail.SingleLine "Zj positive does " & IIf(existsZjPositive, "", "NOT ") & "exist."
        
    Loop
    
'---evaluate result:
    Select Case Result.Status
    
    Case EnumStatus.Optimal
    
        ModuleGauss.CalculateOptimalSolution Result, Definition, tableau, BVS
        
        noteDetail.PrintTestResult
        noteDetail.PrintSummary
    
    Case Else
    
        noteDetail.SingleLine "Status has changed: " & Result.StatusName
        noteDetail.SingleLine "Execution stopped.", , 1, 1, "="
    
    End Select
    
    noteDetail.CloseOutputFile
    
    Exit Sub

errorExecuteSimplex:

    noteDetail.CloseOutputFile
    
    Err.Raise Err.Number, , "Execute Std. Simplex method error." & _
                            NewLine(Err.Description)
    
End Sub

'
'   (A) set dimensions and default values
'
Private Sub InitiateSimplex()
    
    Dim i As Long, j As Long
    
    '   (a) initiate global method values
    stopExecuting = False
'''    iterationCount = 0
    Inc.Reset
    
    existsZjPositive = False
    biggestZjColumnIndex = 0
    
    '   (b) read Objective, s and m
    minMax = IIf(Definition.Objective = "MAX", -1, 1): Inc.Decis
    M = Definition.M
    S = Definition.S
    
    '   (c) calculate D and U
    D = 0
    U = 0
    
    Inc.Loops M
    For i = 1 To M
    
        D = IIf(Definition.ConstraintType(i) <> "=", D + 1, D): Inc.Decis
        U = IIf(Definition.ConstraintType(i) <> "<", U + 1, U): Inc.Decis
    
    Next i

    '   (d) dimension tableau, set initial values to A or zero
    ReDim tableau(1 To M + 1, 1 To S + D + U + 1)
    
    Inc.Loops M * S
    For i = 1 To M
        
        For j = 1 To S
            
            tableau(i, j) = Definition.A(i, j)
        
        Next j
    
    Next i

    '   (e) dimension BVS, set initial values to zero
    ReDim BVS(1 To M)

    '   (f) calculate BIG M
    bigM = 0
    
    Inc.Loops S
    Inc.Decis S
    For i = 1 To S
        
        bigM = IIf(Definition.Cj(i) > bigM, Definition.Cj(i), bigM)
    
    Next i
    
    bigM = 10 * bigM: Inc.MulDiv

    '   (g) dimension simplexCj (expand original Cj with surplus/slack and artificial variables)
    ReDim simplexCj(S + D + U + 1)
    
    Inc.Loops S
    For j = 1 To S
        
        simplexCj(j) = Definition.Cj(j):
        
    Next j

End Sub

'
'   (B) Construct initial tableau for Simplex
'
Private Sub ConstructInitialTableau()

    Dim i As Long
    Dim Dj As Long: Dj = 0
    Dim Uj As Long: Uj = 0

    '   (a) create slack, surplus and artificial
    Inc.Loops M
    Inc.Decis M
    For i = 1 To M
        
        Select Case Definition.ConstraintType(i)
        
        Case "="
        
            Uj = Uj + 1
            tableau(i, S + D + Uj) = 1
            simplexCj(S + D + Uj) = minMax * bigM
            BVS(i) = S + D + Uj
            Inc.MulDiv
            Inc.AddSub 3
        
        Case "<"
        
            Dj = Dj + 1
            tableau(i, S + Dj) = 1
            BVS(i) = S + Dj
            Inc.AddSub 2
        
        Case ">"
        
            Uj = Uj + 1
            Dj = Dj + 1
            tableau(i, S + Dj) = -1
            simplexCj(S + D + Uj) = minMax * bigM
            tableau(i, S + D + Uj) = 1
            BVS(i) = S + D + Uj
            Inc.AddSub 4
            Inc.MulDiv
            
        End Select
        
    Next i

    '   (b) read RHS
    Inc.Loops M
    For i = 1 To M
        
        tableau(i, S + D + U + 1) = Definition.RHS(i)
    
    Next i

    '   (c) calculate z(j)-c(j)
    CalculateZ
    
    '   if Zj positive doesn't exist stop executing
    stopExecuting = Not existsZjPositive

End Sub

'
'   calculate new incoming and outgoing vector, pivot tableau
'
Private Sub CalculateNewTableau()

    '   find best k
    k = biggestZjColumnIndex
    
    Result.Status = NotCalculated
    
    stopExecuting = True
    
    '   calculate r with C/R test
    r = ModuleGauss.ColumnRatioTest(Inc, tableau, k, degeneracyOccured)
    
    '   what if column ratio test doesn't give a result
    Inc.Decis
    If r = 0 Then
        
        Result.Status = EnumStatus.Unbounded
    End If
    
    '   increase iteration count and check for boundaries
    Inc.Decis
    If ModuleGauss.IterationCountExceedsMaxAllowed(Inc) Then
    
        Result.Status = MaxIterationsExceeded
    End If
    
    '   something went wrong
    Inc.Decis
    If Result.Status <> NotCalculated Then Exit Sub
        
    ModuleGauss.GaussPivot Inc, tableau, k, r, BVS
    
    CalculateZ
    
    '   check for positive Zj
    stopExecuting = Not existsZjPositive
    
End Sub

'
'   calculating last row; alter for maximum
'
Private Function CalculateZ()

    Dim i As Long
    Dim j As Long
    Dim tempZ As Double
    Dim tempBiggestZj As Double: tempBiggestZj = 0

    existsZjPositive = False
    biggestZjColumnIndex = 0
    
    '   (a) calculate last row
    For j = 1 To S + D + U + 1
        
        tempZ = 0
        
        Inc.Loops M
        For i = 1 To M
            
            '   calculate cumulative value
            tempZ = tempZ + simplexCj(BVS(i)) * tableau(i, j)
            Inc.AddSub
            Inc.MulDiv
        Next i
        
        Inc.Decis
        If j < S + D + U + 1 Then
            
            '   calculate last row in tableau
            tableau(M + 1, j) = minMax * (tempZ - simplexCj(j))
            Inc.MulDiv
        
            '   check for positive value
            existsZjPositive = IIf(tableau(M + 1, j) > ModuleGauss.eta, True, existsZjPositive)
            Inc.Decis
            
            '   check for biggest value
            Inc.Decis
            If tableau(M + 1, j) > tempBiggestZj Then
            
                tempBiggestZj = tableau(M + 1, j)
                biggestZjColumnIndex = j
            
            End If
        
        Else
            
            '   current objective function value
            tableau(M + 1, j) = tempZ
        
        End If
    
    Next j

    Inc.Decis
    If Not existsZjPositive Then Result.Status = Optimal

End Function

